import base64
import json
import os
import re
from io import StringIO, BytesIO
from urllib.error import URLError
from urllib.request import BaseHandler
from urllib.response import addinfourl

from kubernetes import client, config


K8S_URL_PATTERN = re.compile(
    r'k8s://%(ns)s/%(type)s/%(name)s(/%(key)s*)?' % {
        "ns":   r"(?P<ns>[\.\w-]*)",
        "type": r"(?P<type>\w+)",
        "name": r"(?P<name>[\w-]+)",
        "key":  r"(?P<key>[^\/]+)"
    }
)


class _ParsedUrl:
    def __init__(self, url):
        self.url = url
        m = K8S_URL_PATTERN.match(url)
        if not m:
            raise URLError('url {} is not parsable'.format(url))
        self.resource_type = 'configmap' if m.group('type') in 'cm' else m.group('type')
        self.resource_name = m.group('name')
        self.namespace = 'default' if m.group('ns') in ['', '.'] else m.group('ns')
        self.key = m.group('key')


class K8SHandler(BaseHandler):
    @staticmethod
    def get_api_client():
        if os.getenv('KUBERNETES_SERVICE_HOST'):
            config.load_incluster_config()
        else:
            config.load_kube_config()

        return client.CoreV1Api()

    def __init__(self):
        self.api = self.get_api_client()

    def k8s_open(self, req):
        url = req.get_full_url()
        pu = _ParsedUrl(url)

        if pu.resource_type == 'secret':
            stream = self._read_secret(pu)
        elif pu.resource_type == 'configmap':
            stream = self._read_configmap(pu)
        else:
            raise URLError('url {} is not parsable'.format(pu.url))

        return addinfourl(stream, [], url)

    def _read_configmap(self, pu: _ParsedUrl):
        resource = self.api.read_namespaced_config_map(pu.resource_name, pu.namespace)
        if pu.key:
            content = resource.data[pu.key]
            return StringIO(content)
        else:
            return StringIO(json.dumps(resource.data))

    def _read_secret(self, pu: _ParsedUrl):
        resource = self.api.read_namespaced_secret(pu.resource_name, pu.namespace)
        if pu.key:
            content = resource.data[pu.key]
            return BytesIO(base64.b64decode(content.encode('utf8')))
        else:
            content = {
                k: base64.b64decode(v.encode('utf8')).decode('utf8')
                for k, v in resource.data.items()
            }
            return StringIO(json.dumps(content))



import copy
import logging

from kebab.exceptions import KebabException

_logger = logging.getLogger(__name__)


def update_recursively(value1, value2):
    """
    Update dictionary recursively.
    If they are both dicts, then take the update the keys recursively (Python extend(...) function).
    Otherwise, take the second value.

    :param str|int|float|list|dict|None value1:
    :param str|int|float|list|dict|None value2:
    :return:
    """
    if value2 is None:
        return copy.deepcopy(value1)
    elif not isinstance(value1, dict) or not isinstance(value2, dict):
        # If either d1 or d2 is not dict, then use second one.
        return copy.deepcopy(value2)
    else:
        result = copy.deepcopy(value1)
        for key in value2:
            if key in result:
                result[key] = update_recursively(result[key], value2[key])
            elif value2[key] is not None:
                result[key] = value2[key]
        return result


def fill_recursively(dictionary, key, value, delimiter='.'):
    if delimiter not in key:
        dictionary[key] = value
    else:
        next_key, remain_path = key.split(delimiter, 1)
        inner_dictionary = dictionary.setdefault(next_key, {})
        if isinstance(inner_dictionary, dict):
            try:
                fill_recursively(inner_dictionary,
                                 key=remain_path,
                                 value=value,
                                 delimiter=delimiter)
            except KebabException:
                raise KebabException(f"unable to inflate key {key}, not a dictionary")
        else:
            raise KebabException(f"unable to inflate key {key}, not a dictionary")


def lookup_recursively(dictionary, key, default_value=None, delimiter='.'):
    """
    If there is no delimiter, this is the last level in the path, return directly.

    If there are more delimiters, use the first part of the key to retrieve the next level of
    the dictionary and lookup recursively.

    Unless the following two situation happens:
        1. the value of current level doesn't exist
        2. the value of current level is not a dict

    :param dict dictionary:
    :param str key:
    :param Any default_value:
    :param str delimiter:
    """
    if delimiter not in key:
        # no more split, value supposed to be in this level
        return dictionary.get(key, default_value)
    else:
        next_key, remain_path = key.split(delimiter, 1)
        # the value of current level has to exist
        if next_key in dictionary:
            inner_dictionary = dictionary.get(next_key)
            # the value of current level has to be a dict
            if isinstance(inner_dictionary, dict):
                # enter next level
                return lookup_recursively(inner_dictionary,
                                          key=remain_path,
                                          default_value=default_value,
                                          delimiter=delimiter)
        return default_value


def flatten(dictionary, path=None, delimiter='_'):
    """
    Given the dictionary, flatten it to one level (separated by underscore).
    :param dict[str, any] dictionary: the dictionary to flatten
    :param list[str] path: the path to the current dictionary
    :param delimiter: the delimiter to join hierarchical keys.
    :return: flattened dictionary
    """
    path = path or []

    flattened = {}
    for child_key, child_value in dictionary.items():
        child_path = path + [child_key]
        if isinstance(child_value, dict):
            flattened.update(flatten(child_value, child_path))
        else:
            # child_value could be either a list or str, int, unicode, float, ...
            flattened[delimiter.join(child_path)] = child_value
    return flattened



from kebab.sources import KebabSource


class Field:
    def __init__(self, config_name='.', default_value=None, required=False, expected_type=None):
        self.config_name = config_name
        self.default_value = default_value
        self.required = required
        self.expected_type = expected_type


def kebab_config(auto_repr=False):
    if not isinstance(auto_repr, bool):
        # Most likely auto_repr here is the klass when @kebab_config is used without parameters
        return kebab_config()(auto_repr)

    def _make_kebab_config_class(klass):
        if hasattr(klass, '__init__'):
            user_init = klass.__init__
        else:
            user_init = None

        def __init__(self, source: KebabSource, *args, **kwargs):
            if isinstance(source, KebabSource):
                for field_name, field in vars(klass).items():
                    if isinstance(field, Field):
                        setattr(self, field_name, source.get(**field.__dict__))
            if user_init:
                user_init(self, *args, **kwargs)
        setattr(klass, '__init__', __init__)

        if auto_repr:
            def __repr__(self):
                result = []
                for key, value in vars(self).items():
                    try:
                        if not key.startswith("__") and not key.endswith("__"):
                            #     value = getattr(self, key)
                            result.append(f'{key}: {value!r}')
                    except AttributeError:
                        continue
                return f'{self.__class__.__name__}({", ".join(result)})'
            setattr(klass, '__repr__', __repr__)

        return klass
    return _make_kebab_config_class


class KebabConfigMeta(type):
    def __new__(mcs, name, bases, namespace, **kwargs):
        new_cls = super(KebabConfigMeta, mcs).__new__(mcs, name, bases, namespace)
        return kebab_config()(new_cls)



import threading
from urllib.error import URLError
from urllib.request import BaseHandler

import oss2
from alibabacloud import ClientConfig
from alibabacloud.credentials.provider import DefaultChainedCredentialsProvider

DEFAULT_OSS_ENDPOINT = "https://oss-cn-hangzhou.aliyuncs.com"


class OSSHandler(BaseHandler):
    def __init__(self, credential_provider=None, client_config=None,
                 resolve_endpoint=True, prefer_internal_point=False):
        self._client_config = client_config or ClientConfig()
        self._cred_provider = credential_provider or DefaultChainedCredentialsProvider(client_config=client_config)
        self._auth = self._get_auth()
        self._resolve_endpoint = resolve_endpoint
        self._prefer_internal_point = prefer_internal_point
        self._bucket_info_cache_lock = threading.RLock()
        self._bucket_info_cache = {}

        if not self._client_config.endpoint:
            self._client_config.endpoint = DEFAULT_OSS_ENDPOINT

    def _get_auth(self):
        cred = self._cred_provider.provide()
        return oss2.Auth(cred.access_key_id, cred.access_key_secret)

    def _resolve_bucket_endpoint(self, bucket_name):
        if not self._resolve_endpoint:
            return self._client_config.endpoint

        # lookup endpoint in bucket_info, choose the endpoint and normalize
        with self._bucket_info_cache_lock:
            if bucket_name not in self._bucket_info_cache:
                bucket = oss2.Bucket(self._auth, self._client_config.endpoint, bucket_name)
                bucket_info = bucket.get_bucket_info()
                self._bucket_info_cache[bucket_name] = bucket_info

        bucket_info = self._bucket_info_cache[bucket_name]
        return self._normalize_endpoint(bucket_info.internal_endpoint if self._prefer_internal_point else
                                        bucket_info.extranet_endpoint)

    @staticmethod
    def _normalize_endpoint(endpoint):
        """
        add https:// scheme if not present
        :param endpoint: endpoint str with or without scheme
        :return:
        """
        endpoint = endpoint.strip()
        if not endpoint.startswith('http://') and not endpoint.startswith('https://'):
            return 'https://' + endpoint
        else:
            return endpoint

    def oss_open(self, req):
        try:
            selector = req.selector
        except AttributeError:
            selector = req.get_selector()
        key_name = selector.lstrip('/')
        bucket_name = req.host

        if key_name is None:
            raise URLError('no such resource: {}'.format(req.get_full_url()))
        if not bucket_name or not key_name:
            raise URLError('url must be in the format oss://<bucket>/<key>')

        endpoint = self._resolve_bucket_endpoint(bucket_name)
        bucket = oss2.Bucket(self._auth, endpoint, bucket_name)
        return bucket.get_object(key_name)



import json
import time

import click

from kebab import load_source
from kebab.sources import _DISABLE_RELOAD

DISABLE_RELOAD = _DISABLE_RELOAD


@click.command()
@click.argument('sources', nargs=-1, type=str)
@click.option('--env-vars/--no-env-vars', default=False)
@click.option('-k', '--key', default='.', type=str)
@click.option('-w', '--watch', default=DISABLE_RELOAD, type=float)
def run(sources, key, watch, env_vars):
    if not sources:
        sources = "app.yaml"
    conf = load_source(default_urls=sources, include_env_var=env_vars, reload_interval_in_secs=watch)
    while True:
        click.echo(json.dumps(conf.get(key)))
        if watch == DISABLE_RELOAD:
            break
        time.sleep(watch)


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kebab', 'kebab.cli']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'jmespath>=0.10.0,<0.11.0',
 'pyyaml>=5.1,<6.0',
 'six>=1.14.0,<2.0.0']

extras_require = \
{'ali': ['oss2>=2.11.0,<3.0.0', 'alibaba-cloud-python-sdk-v2>=1.0.6,<2.0.0'],
 'aws': ['boto3>=1.13.12,<2.0.0'],
 'k8s': ['kubernetes>=11.0.0,<12.0.0']}

entry_points = \
{'console_scripts': ['kebab = kebab.cli:run']}

setup_kwargs = {
    'name': 'pykebab',
    'version': '0.2.0',
    'description': '',
    'long_description': None,
    'author': 'Yangming Huang',
    'author_email': 'leonmax@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from kebab.sources import (
    DictSource, EnvVarSource, KebabSource, UnionSource, UrlSource,
    literal, union, load_source, default_source
)
from kebab.magic import kebab_config, Field

__all__ = [
    "DictSource", "EnvVarSource", "KebabSource", "UnionSource", "UrlSource", "Field",
    "kebab_config", "literal", "union", "load_source", "default_source"
]
